//********************************************************************************
// hashsep.h
// Interface for separate chaining hash table
//
// N.B. These hash table functions are modified from code by Mark Allen Weiss
// http://www.cs.fiu.edu/~weiss/dsaa_c2e/files.html
//********************************************************************************

#ifndef True
#define True 1
#endif

#ifndef False
#define False 0
#endif

typedef unsigned int Index;

/*
 * A hash table for holding elements of type: unsigned int key, unsigned int value
 */

typedef unsigned int IHTKeyType;
typedef unsigned int IHTValueType;

typedef struct inthashtable_element {
	IHTKeyType key;
	IHTValueType value;
} IHTElementType;

#ifndef _HashSep_H

struct IHTListNode;
typedef struct IHTListNode * IHTPosition;
struct IntHashTbl;
typedef struct IntHashTbl * IntHashTable;

IntHashTable InitializeIntHashTable( int TableSize );
void ClearIntHashTable( IntHashTable H ); // added by TJT
void DestroyIntHashTable( IntHashTable H );
IHTPosition FindInIntHashTable( IHTKeyType Key, IntHashTable H );
int InsertInIntHashTable( IHTElementType Elem, IntHashTable H );
void RemoveFromIntHashTable( IHTKeyType Key, IntHashTable H ); // added by TJT
IHTElementType RetrieveFromIntHashTable( IHTPosition P );

#endif /* _HashSep_H */

/*
 * A hash table for holding elements of type: unsigned int key, void* value
 */

typedef unsigned int PHTKeyType;
typedef void* PHTValueType;

typedef struct ptrhashtable_element {
	PHTKeyType key;
	PHTValueType value;
} PHTElementType;

#ifndef _HashSep_H

struct PHTListNode;
typedef struct PHTListNode * PHTPosition;
struct PtrHashTbl;
typedef struct PtrHashTbl * PtrHashTable;

PtrHashTable InitializePtrHashTable( int TableSize );
void ClearPtrHashTable( PtrHashTable H ); // added by TJT
void DestroyPtrHashTable( PtrHashTable H );
PHTPosition FindInPtrHashTable( PHTKeyType Key, PtrHashTable H );
void InsertInPtrHashTable( PHTElementType Elem, PtrHashTable H );
PHTElementType RetrieveFromPtrHashTable( PHTPosition P );

#endif  /* _HashSep_H */


//********************************************************************************
// The following code would normally reside in a separate file, hashsep.c
// It is included here because CAMELot has a limited capacity for dealing with
// multiple source files
//
// N.B. These hash table functions are modified from code by Mark Allen Weiss
// http://www.cs.fiu.edu/~weiss/dsaa_c2e/files.html
//********************************************************************************

#include <stdlib.h>
#include <stdio.h>

#define MinTableSize (10)

void error( char * msg /*, char * var*/ );
void fatalError( char * msg /*, char * var*/ );

struct IHTListNode
{
	IHTElementType Element;
	IHTPosition    Next;
};

typedef IHTPosition IHTList;

/* List *TheList will be an array of lists, allocated later */
/* The lists use headers (for simplicity), */
/* though this wastes space */
struct IntHashTbl
{
	int TableSize;
	int NumEntries;
	IHTList *TheLists;
};

struct PHTListNode
{
	PHTElementType Element;
	PHTPosition    Next;
};

typedef PHTPosition PHTList;

/* List *TheList will be an array of lists, allocated later */
/* The lists use headers (for simplicity), */
/* though this wastes space */
struct PtrHashTbl
{
	int TableSize;
	PHTList *TheLists;
};


void error(char * msg /*, char * var*/)
{
	fprintf(stderr, msg /*, var*/);
	fflush(stderr);
}


/*
void fatalError(char * msg /*, char * var/)
{
	fprintf(stderr, msg /*, var/);
	fflush(stderr);
	exit(EXIT_FAILURE);
}
*/


/* Return next prime; assume N >= 10 */
static int NextPrime( int N )
{
	int i;

	if( N % 2 == 0 )
		N++;
	for( ; ; N += 2 )
	{
		for( i = 3; i * i <= N; i += 2 )
			if( N % i == 0 )
				goto ContOuter;  /* Sorry about this! */
		return N;
	  ContOuter: ;
	}
}

		
/* Hash function for ints */
Index IntHash( IHTKeyType Key, int TableSize )
{
	return Key % TableSize;
}

		
IntHashTable InitializeIntHashTable( int TableSize )
{
	IntHashTable H;
    int i;

	if( TableSize < MinTableSize )
    {
		error( "Table size too small" );
		return NULL;
	}

	/* Allocate table */
	H = malloc( sizeof( struct IntHashTbl ) );
	if( H == NULL )
	{
		fatalError( "Out of space!!!" );
	}

	H->TableSize = NextPrime( TableSize );

	/* Allocate array of lists */
	H->TheLists = malloc( sizeof( IHTList ) * H->TableSize );
	if( H->TheLists == NULL )
	{
		fatalError( "Out of space!!!" );
	}

	/* Allocate list headers */
	for( i = 0; i < H->TableSize; i++ )
	{
		H->TheLists[ i ] = malloc( sizeof( struct IHTListNode ) );
		if( H->TheLists[ i ] == NULL )
		{
			fatalError( "Out of space!!!" );
		}
        else
			H->TheLists[ i ]->Next = NULL;
	}
	
	H->NumEntries = 0;

	return H;
}


IHTPosition FindInIntHashTable( IHTKeyType Key, IntHashTable H )
{
	IHTPosition P;
    IHTList L;

    L = H->TheLists[ IntHash( Key, H->TableSize ) ];
    P = L->Next;
    while( P != NULL && P->Element.key != Key )
                     /* Probably need strcmp!! */
		P = P->Next;
	return P;
}

// modified by TJT to return True on successful insertion, False otherwise
int InsertInIntHashTable( IHTElementType Elem, IntHashTable H )
{
	IHTPosition Pos, NewCell;
    IHTList L;

    Pos = FindInIntHashTable( Elem.key, H );
    if ( Pos == NULL )   /* Key is not found */
    {
		NewCell = malloc( sizeof( struct IHTListNode ) );
        if( NewCell == NULL )
		{
			fatalError( "Out of space!!!" );
		}
        else
        {
			L = H->TheLists[ IntHash( Elem.key, H->TableSize ) ];
            NewCell->Next = L->Next;
            NewCell->Element = Elem;  /* Probably need strcpy! */
            L->Next = NewCell;
			H->NumEntries++;
		}
		
		return True;
	}
	else
		return False;
}


// This is an extra function added by TJT
void RemoveFromIntHashTable( IHTKeyType Key, IntHashTable H )
{
	IHTPosition P, PrevP;
    IHTList L;

    L = H->TheLists[ IntHash( Key, H->TableSize ) ];
	PrevP = L;
    P = L->Next;
    while( P != NULL && P->Element.key != Key ) /* Probably need strcmp!! */
	{
		PrevP = PrevP->Next;
		P = P->Next;
	}
	
	if ( P != NULL )
	{
		PrevP->Next = P->Next;
		free( P );
		H->NumEntries--;
	}
}


IHTElementType RetrieveFromIntHashTable( IHTPosition P )
{
	return P->Element;
}


// This is an extra function added by TJT
void ClearIntHashTable( IntHashTable H ) 
{
	int i;

	for( i = 0; i < H->TableSize; i++ )
	{
		IHTList L = H->TheLists[ i ];
		IHTPosition P = L->Next;
		IHTPosition Tmp;
		
		while( P != NULL )
		{
			Tmp = P->Next;
			free( P );
			P = Tmp;
		}
		
		L->Next = NULL;
	}
	
	H->NumEntries = 0;
}


void DestroyIntHashTable( IntHashTable H )
{
	int i;

	for( i = 0; i < H->TableSize; i++ )
	{
		IHTPosition P = H->TheLists[ i ];
		IHTPosition Tmp;

		while( P != NULL )
		{
			Tmp = P->Next;
			free( P );
			P = Tmp;
		}
	}

	free( H->TheLists );
	free( H );
}


/* Following are the same routines defined for PtrHashTables */
		
PtrHashTable InitializePtrHashTable( int TableSize )
{
	PtrHashTable H;
    int i;

	if( TableSize < MinTableSize )
    {
		error( "Table size too small" );
		return NULL;
	}

	/* Allocate table */
	H = malloc( sizeof( struct PtrHashTbl ) );
	if( H == NULL )
	{
		fatalError( "Out of space!!!" );
	}

	H->TableSize = NextPrime( TableSize );

	/* Allocate array of lists */
	H->TheLists = malloc( sizeof( PHTList ) * H->TableSize );
	if( H->TheLists == NULL )
	{
		fatalError( "Out of space!!!" );
	}

	/* Allocate list headers */
	for( i = 0; i < H->TableSize; i++ )
	{
		H->TheLists[ i ] = malloc( sizeof( struct PHTListNode ) );
		if( H->TheLists[ i ] == NULL )
		{
			fatalError( "Out of space!!!" );
		}
        else
			H->TheLists[ i ]->Next = NULL;
	}

	return H;
}


PHTPosition FindInPtrHashTable( PHTKeyType Key, PtrHashTable H )
{
	PHTPosition P;
    PHTList L;

    L = H->TheLists[ IntHash( Key, H->TableSize ) ];
    P = L->Next;
    while( P != NULL && P->Element.key != Key )
                     /* Probably need strcmp!! */
		P = P->Next;
	return P;
}


void InsertInPtrHashTable( PHTElementType Elem, PtrHashTable H )
{
	PHTPosition Pos, NewCell;
    PHTList L;

    Pos = FindInPtrHashTable( Elem.key, H );
    if( Pos == NULL )   /* Key is not found */
    {
		NewCell = malloc( sizeof( struct PHTListNode ) );
        if( NewCell == NULL )
		{
			fatalError( "Out of space!!!" );
		}
        else
        {
			L = H->TheLists[ IntHash( Elem.key, H->TableSize ) ];
            NewCell->Next = L->Next;
            NewCell->Element = Elem;  /* Probably need strcpy! */
            L->Next = NewCell;
		}
	}
}


PHTElementType RetrieveFromPtrHashTable( PHTPosition P )
{
	return P->Element;
}


// This is an extra function added by TJT
void ClearPtrHashTable( PtrHashTable H ) 
{
	// N.B. This does NOT free the memory associated with element.value for each entry
	
	int i;

	for( i = 0; i < H->TableSize; i++ )
	{
		PHTList L = H->TheLists[ i ];
		PHTPosition P = L->Next;
		PHTPosition Tmp;
		
		while( P != NULL )
		{
			Tmp = P->Next;
			free( P );
			P = Tmp;
		}
		
		L->Next = NULL;
	}
}


void DestroyPtrHashTable( PtrHashTable H )
{
	int i;

	for( i = 0; i < H->TableSize; i++ )
	{
		PHTPosition P = H->TheLists[ i ];
		PHTPosition Tmp;

		while( P != NULL )
		{
			Tmp = P->Next;
			free( P );
			P = Tmp;
		}
	}

	free( H->TheLists );
	free( H );
}


